package de.citunius.businesslogic.hellochatscript.services;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.lang.invoke.MethodHandles;
import java.net.Socket;
import java.util.HashMap;

import org.apache.log4j.Logger;

public class ChatScriptService {

	String host = null;
	int port = 1024; // Default port
	String message = null;
	String botname = null;
	String username = null;
	final char nullChar = (char) 0;
	boolean initialized = false;
	static Logger logger = Logger.getLogger(MethodHandles.lookup().lookupClass());

	public ChatScriptService(String ihost, int port, String ibotname, String iusername) {
		this.host = ihost;
		this.port = port;
		this.botname = ibotname;
		this.username = iusername;
	}

	/**
	 * Send a chat message to ChatScript
	 * 
	 * @param input		the message text
	 * @return	the ChatScript reply
	 */
	public String sendChat(String input) {
		if (!initialized)
			init(null);
		return doMessage(this.username + nullChar + nullChar + input + nullChar);
	}

	/**
	 * Send a message to ChatScript
	 * 
	 * <p>Open a socket to ChatScript server and process message</p> 
	 * 
	 * @param message		the message text
	 * @return	the ChatScript reply
	 */
	private String doMessage(String message) {
		Socket echoSocket;
		String response = null;
		logger.info("Opening socket to host ["+host+"] on port ["+port+"]");
		try {			
			echoSocket = new Socket(this.host, this.port);
			PrintWriter out = new PrintWriter(echoSocket.getOutputStream(), true);
			BufferedReader in = new BufferedReader(new InputStreamReader(echoSocket.getInputStream()));
			out.println(message);
			response = in.readLine();
			echoSocket.close();
		} catch (IOException e) {
			logger.error("Error: " + e.getMessage());
			return null;
		}
		return response;

	}

	/**
	 * Init ChatScript
	 * 
	 * @param args	the arguments
	 */
	public void init(HashMap<String, String> args) {
		logger.info("ChatScript Bot init()...");
		logger.info(" Got back: " + doMessage(nullChar + "1" + nullChar + nullChar));
		logger.info(" Now starting conversation..." + doMessage(this.username + nullChar + nullChar + nullChar));
		logger.info(" Conversation started with default bot as " + this.username);
		initialized = true;
	}

	/**
	 * Get Bot type
	 * 
	 * @return the bot type
	 */
	public String getBotType() {
		return "ChatScript";
	}

	/**
	 * Reset
	 * 
	 * @return the bot type
	 */
	public void reset() {
		doMessage(this.username + nullChar + nullChar + ":reset" + nullChar);
	}

}