package test.de.citunius.chatscript;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.HashMap;
import java.util.Scanner;

public class ChatScriptBotDialogueTestMain {

	String chatScriptServer = "localhost";
	int chatScriptServerPort = 1024;
	
	char nullChar = (char) 0;
	String userName = "Jerry"; // or take as input

	public String processInput(String input) {
		return doMessage("Varun" + nullChar + nullChar + input + nullChar);
	}

	/**
	 * Send a message to ChatScript
	 * 
	 * @param message		the message text
	 * @return	the ChatScript reply
	 */
	private String doMessage(String message) {
		Socket echoSocket;
		String resp = "";
		try {
			echoSocket = new Socket(chatScriptServer, chatScriptServerPort);
			PrintWriter out = new PrintWriter(echoSocket.getOutputStream(), true);
			BufferedReader in = new BufferedReader(new InputStreamReader(echoSocket.getInputStream()));
			out.println(message);
			resp = in.readLine();
			echoSocket.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Error: " + e.getMessage());
		}
		return resp;
	}

	/**
	 * Init ChatScript
	 * 
	 * @param args	the arguments
	 */
	public void init(HashMap<String, String> args) {
		System.out.println("ChatScript Bot init()...");
		System.out.println(" Got back: " + doMessage(nullChar + "1" + nullChar + nullChar));
		System.out.println(" Now starting conversation..." + doMessage(userName + nullChar + nullChar + nullChar));
		System.out.println(" Conversation started with default bot as " + userName);
	}

	/**
	 * Get Bot type
	 * 
	 * @return the bot type
	 */
	public String getBotType() {
		return "ChatScript";
	}

	/**
	 * Main
	 * 
	 * @param arg	the argument
	 */
	public static void main(String[] arg) {
		ChatScriptBotDialogueTestMain t = new ChatScriptBotDialogueTestMain();
		t.init(null);
		Scanner in = new Scanner(System.in);
		while (true) {
			String input = in.nextLine();
			System.out.println("You: "+input);
			if (input.equals("bye")) {
				// Send last query to ChatScript, then exit program
				System.out.println("ChatScript: "+t.processInput(input));
				System.out.println("ChatScript session ended.");
				in.close();
				return;
			} else {
				// Process user message to ChatScript
				System.out.println("ChatScript: "+t.processInput(input));
			}			
		}
	}
}